// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2017 and onwards Google, Inc.

#include <baumwelch/trainscript.h>
#include <fst/script/script-impl.h>

namespace fst {
namespace script {

bool GetExpectationTableType(const string &str, ExpectationTableType *etype) {
  if (str == "global") {
    *etype = GLOBAL;
  } else if (str == "state") {
    *etype = STATE;
  } else if (str == "ilabel") {
    *etype = ILABEL;
  } else if (str == "state_ilabel") {
    *etype = STATE_ILABEL;
  } else {
    return false;
  }
  return true;
}

// 1: Pair construction.
bool TrainBaumWelch(FarReaderClass *plaintext, FarReaderClass *ciphertext,
                    MutableFstClass *channel, ExpectationTableType etype,
                    const BaumWelchTrainOptions &opts) {
  if (!internal::ArcTypesMatch(*plaintext, *channel, "TrainBaumWelch") ||
      !internal::ArcTypesMatch(*ciphertext, *channel, "TrainBaumWelch")) {
    channel->SetProperties(kError, kError);
    return false;
  }
  TrainBaumWelchInnerArgs1 iargs(plaintext, ciphertext, channel, etype, opts);
  TrainBaumWelchArgs1 args(iargs);
  Apply<Operation<TrainBaumWelchArgs1>>("TrainBaumWelch", channel->ArcType(),
                                        &args);
  return args.retval;
}

// 2: Decipherment construction.
bool TrainBaumWelch(const FstClass &plaintext, FarReaderClass *ciphertext,
                    MutableFstClass *channel, ExpectationTableType etype,
                    const BaumWelchTrainOptions &opts) {
  if (!internal::ArcTypesMatch(plaintext, *channel, "TrainBaumWelch") ||
      !internal::ArcTypesMatch(*ciphertext, *channel, "TrainBaumWelch")) {
    channel->SetProperties(kError, kError);
    return false;
  }
  TrainBaumWelchInnerArgs2 iargs(plaintext, ciphertext, channel, etype, opts);
  TrainBaumWelchArgs2 args(iargs);
  Apply<Operation<TrainBaumWelchArgs2>>("TrainBaumWelch", channel->ArcType(),
                                        &args);
  return args.retval;
}

REGISTER_FST_OPERATION(TrainBaumWelch, StdArc, TrainBaumWelchArgs1);
REGISTER_FST_OPERATION(TrainBaumWelch, LogArc, TrainBaumWelchArgs1);
REGISTER_FST_OPERATION(TrainBaumWelch, Log64Arc, TrainBaumWelchArgs1);

REGISTER_FST_OPERATION(TrainBaumWelch, StdArc, TrainBaumWelchArgs2);
REGISTER_FST_OPERATION(TrainBaumWelch, LogArc, TrainBaumWelchArgs2);
REGISTER_FST_OPERATION(TrainBaumWelch, Log64Arc, TrainBaumWelchArgs2);

}  // namespace script
}  // namespace fst

